@@ -1,5 +1,6 @@
 # -*- coding: utf-8 -*-
 
+import math
 import os
 import re
 import sys
@@ -18,7 +19,8 @@ from options import debug
 from util_list import floatlist, strlist
 from util_str import safe_str, safe_unicode
 from wxaddons import wx
-from wxwindows import BaseApp, ConfirmDialog, InfoDialog, InvincibleFrame
+from wxwindows import (BaseApp, BitmapBackgroundPanel, ConfirmDialog, InfoDialog,
+					   InvincibleFrame)
 from wxfixes import GenBitmapButton as BitmapButton
 try:
 	import RealDisplaySizeMM as RDSMM
@@ -125,86 +127,128 @@ class MeasureFrame(InvincibleFrame):
 								 name="measureframe")
 		self.SetIcons(config.get_icon_bundle([256, 48, 32, 16], appname))
 		self.Bind(wx.EVT_CLOSE, self.close_handler, self)
-		self.Bind(wx.EVT_MOVE, self.move_handler, self)
+		if os.getenv("XDG_SESSION_TYPE") != "wayland":
+			self.Bind(wx.EVT_MOVE, self.move_handler, self)
+		self.Bind(wx.EVT_SHOW, self.show_handler)
+		self.Bind(wx.EVT_KILL_FOCUS, self.focus_lost_handler)
 		self.Bind(wx.EVT_SET_FOCUS, self.focus_handler)
-		self.panel = wx.Panel(self, -1)
-		self.sizer = wx.GridSizer(3, 1, 0, 0)
+		self.panel = BitmapBackgroundPanel(self)
+		self.panel.scalebitmap = (False, False)
+
+		if os.getenv("XDG_SESSION_TYPE") == "wayland":
+			self.sizer = wx.FlexGridSizer(3, 1, 0, 0)
+			self.sizer.AddGrowableCol(0)
+			self.sizer.AddGrowableRow(1)
+		else:
+			self.sizer = wx.GridSizer(3, 1, 0, 0)
 		self.panel.SetSizer(self.sizer)
 
 		self.hsizer = wx.BoxSizer(wx.HORIZONTAL)
-		self.sizer.Add(self.hsizer, flag=wx.ALIGN_CENTER_HORIZONTAL | wx.ALL | 
-										 wx.ALIGN_TOP, border=10)
+		self.sizer.Add(self.hsizer, flag=wx.ALIGN_CENTER_HORIZONTAL | 
+										 wx.ALIGN_TOP)
 
 		self.zoommaxbutton = BitmapButton(self.panel, -1, 
-										  geticon(32, "zoom-best-fit"), 
-										  style=wx.NO_BORDER)
+										  geticon(16, "zoom-best-fit"), 
+										  style=wx.NO_BORDER,
+										  name="zoommaxbutton")
+		self.zoommaxbutton.Bind(wx.EVT_KILL_FOCUS, self.focus_lost_handler)
 		self.Bind(wx.EVT_BUTTON, self.zoommax_handler, self.zoommaxbutton)
-		self.hsizer.Add(self.zoommaxbutton, flag=wx.ALIGN_CENTER)
+		self.hsizer.Add(self.zoommaxbutton, flag=wx.ALIGN_CENTER | wx.LEFT |
+												 wx.TOP | wx.BOTTOM, border=10)
 		self.zoommaxbutton.SetToolTipString(lang.getstr("measureframe.zoommax"))
 
-		self.hsizer.Add((2, 1))
+		self.hsizer.Add((8, 0))
 
 		self.zoominbutton = BitmapButton(self.panel, -1, 
-										 geticon(32, "zoom-in"), 
-										 style=wx.NO_BORDER)
+										 geticon(16, "zoom-in"), 
+										 style=wx.NO_BORDER,
+										 name="zoominbutton")
+		self.zoominbutton.Bind(wx.EVT_KILL_FOCUS, self.focus_lost_handler)
 		self.Bind(wx.EVT_BUTTON, self.zoomin_handler, self.zoominbutton)
-		self.hsizer.Add(self.zoominbutton, flag=wx.ALIGN_CENTER)
+		self.hsizer.Add(self.zoominbutton, flag=wx.ALIGN_CENTER | wx.TOP |
+												wx.BOTTOM, border=10)
 		self.zoominbutton.SetToolTipString(lang.getstr("measureframe.zoomin"))
 
-		self.hsizer.Add((2, 1))
+		self.hsizer.Add((8, 0))
 
 		self.zoomnormalbutton = BitmapButton(self.panel, -1, 
-											 geticon(32, "zoom-original"), 
-											 style=wx.NO_BORDER)
+											 geticon(16, "zoom-original"), 
+											 style=wx.NO_BORDER,
+											 name="zoomnormalbutton")
+		self.zoomnormalbutton.Bind(wx.EVT_KILL_FOCUS, self.focus_lost_handler)
 		self.Bind(wx.EVT_BUTTON, self.zoomnormal_handler, self.zoomnormalbutton)
-		self.hsizer.Add(self.zoomnormalbutton, flag=wx.ALIGN_CENTER)
+		self.hsizer.Add(self.zoomnormalbutton, flag=wx.ALIGN_CENTER | wx.TOP |
+												wx.BOTTOM, border=10)
 		self.zoomnormalbutton.SetToolTipString(lang.getstr("measureframe."
 														   "zoomnormal"))
 
-		self.hsizer.Add((2, 1))
+		self.hsizer.Add((8, 0))
 
 		self.zoomoutbutton = BitmapButton(self.panel, -1, 
-										  geticon(32, "zoom-out"), 
-										  style=wx.NO_BORDER)
+										  geticon(16, "zoom-out"), 
+										  style=wx.NO_BORDER,
+										  name="zoomoutbutton")
+		self.zoomoutbutton.Bind(wx.EVT_KILL_FOCUS, self.focus_lost_handler)
 		self.Bind(wx.EVT_BUTTON, self.zoomout_handler, self.zoomoutbutton)
-		self.hsizer.Add(self.zoomoutbutton, flag=wx.ALIGN_CENTER)
+		self.hsizer.Add(self.zoomoutbutton, flag=wx.ALIGN_CENTER | wx.RIGHT |
+												 wx.TOP | wx.BOTTOM, border=10)
 		self.zoomoutbutton.SetToolTipString(lang.getstr("measureframe.zoomout"))
 
-		self.centerbutton = BitmapButton(self.panel, -1, 
-										 geticon(32, "window-center"), 
-										 style=wx.NO_BORDER)
-		self.Bind(wx.EVT_BUTTON, self.center_handler, self.centerbutton)
-		self.sizer.Add(self.centerbutton, flag=wx.ALIGN_CENTER | wx.LEFT | 
-											   wx.RIGHT, border=10)
-		self.centerbutton.SetToolTipString(lang.getstr("measureframe.center"))
+		if os.getenv("XDG_SESSION_TYPE") != "wayland":
+			# No manual centering under Wayland...
+			self.centerbutton = BitmapButton(self.panel, -1, 
+											 geticon(16, "window-center"), 
+											 style=wx.NO_BORDER,
+											 name="centerbutton")
+			self.centerbutton.Bind(wx.EVT_KILL_FOCUS, self.focus_lost_handler)
+			self.Bind(wx.EVT_BUTTON, self.center_handler, self.centerbutton)
+			self.sizer.Add(self.centerbutton, flag=wx.ALIGN_CENTER | wx.LEFT | 
+												   wx.RIGHT, border=10)
+			self.centerbutton.SetToolTipString(lang.getstr("measureframe.center"))
+		else:
+			self.centermsg = wx.StaticText(self.panel, -1,
+										   lang.getstr("measureframe.center.manual"),
+										   style=wx.ALIGN_CENTER_HORIZONTAL)
+			self.centermsg.Wrap(self.hsizer.MinSize[0] - 20)
+			self.sizer.Add(self.centermsg, flag=wx.ALIGN_CENTER | wx.LEFT | 
+												wx.RIGHT, border=10)
 
 		self.vsizer = wx.BoxSizer(wx.VERTICAL)
 		self.sizer.Add(self.vsizer, flag=wx.ALIGN_BOTTOM | 
 										 wx.ALIGN_CENTER_HORIZONTAL)
 
-		self.measure_darken_background_cb = wx.CheckBox(self.panel, -1, 
-			lang.getstr("measure.darken_background"))
-		self.measure_darken_background_cb.SetValue(
-			bool(int(getcfg("measure.darken_background"))))
-		self.Bind(wx.EVT_CHECKBOX, self.measure_darken_background_ctrl_handler, 
-				  id=self.measure_darken_background_cb.GetId())
-		self.vsizer.Add(self.measure_darken_background_cb, 
-						flag=wx.ALIGN_BOTTOM | wx.ALIGN_CENTER_HORIZONTAL | 
-							 wx.LEFT | wx.RIGHT | wx.TOP, border=10)
+		if (os.getenv("XDG_SESSION_TYPE") != "wayland" or
+			config.is_virtual_display()):
+			self.measure_darken_background_cb = wx.CheckBox(self.panel, -1, 
+				lang.getstr("measure.darken_background"))
+			self.measure_darken_background_cb.SetValue(
+				bool(int(getcfg("measure.darken_background"))))
+			self.measure_darken_background_cb.Bind(wx.EVT_KILL_FOCUS,
+												   self.focus_lost_handler)
+			self.Bind(wx.EVT_CHECKBOX, self.measure_darken_background_ctrl_handler, 
+					  id=self.measure_darken_background_cb.GetId())
+			self.vsizer.Add(self.measure_darken_background_cb, 
+							flag=wx.ALIGN_BOTTOM | wx.ALIGN_CENTER_HORIZONTAL | 
+								 wx.LEFT | wx.RIGHT | wx.TOP, border=10)
+		else:
+			self.vsizer.Add((10, 10))
 
 		self.measurebutton = wx.Button(self.panel, -1, 
-			lang.getstr("measureframe.measurebutton"))
+			lang.getstr("measureframe.measurebutton"), name="measurebutton")
+		self.measurebutton.Bind(wx.EVT_KILL_FOCUS, self.focus_lost_handler)
 		self.Bind(wx.EVT_BUTTON, self.measure_handler, self.measurebutton)
 		self.vsizer.Add(self.measurebutton, flag=wx.ALIGN_BOTTOM | 
 												 wx.ALIGN_CENTER_HORIZONTAL | 
 												 wx.ALL, border=10)
 		self.measurebutton.SetMaxFontSize(11)
 		self.measurebutton.SetDefault()
-		self.measurebutton.SetFocus()
+		self.last_focused = self.measurebutton
 
 		self.display_no = wx.Display.GetFromWindow(self)
 		self.display_rects = get_display_rects()
 
+		self.sizer.SetSizeHints(self)
+
 	def measure_darken_background_ctrl_handler(self, event):
 		if self.measure_darken_background_cb.GetValue() and \
 		   getcfg("measure.darken_background.show_warning"):
@@ -246,8 +290,10 @@ class MeasureFrame(InvincibleFrame):
 
 	def Show(self, show=True):
 		if show:
-			self.measure_darken_background_cb.SetValue(
-				bool(int(getcfg("measure.darken_background"))))
+			self.show_controls()
+			if hasattr(self, "measure_darken_background_cb"):
+				self.measure_darken_background_cb.SetValue(
+					bool(int(getcfg("measure.darken_background"))))
 			if self.Parent and hasattr(self.Parent, "display_ctrl"):
 				display_no = self.Parent.display_ctrl.GetSelection()
 			else:
@@ -265,7 +311,16 @@ class MeasureFrame(InvincibleFrame):
 			setcfg("dimensions.measureframe", self.get_dimensions())
 			if self.Parent and hasattr(self.Parent, "get_set_display"):
 				self.Parent.get_set_display()
-		wx.Frame.Show(self, show)
+		if isinstance(self, wx.Dialog):
+			if show:
+				self.ShowModal()
+			else:
+				if self.IsModal():
+					self.EndModal(wx.ID_OK)
+				else:
+					wx.Dialog.Hide(self)
+		else:
+			wx.Frame.Show(self, show)
 
 	def Hide(self):
 		self.Show(False)
@@ -319,10 +374,19 @@ class MeasureFrame(InvincibleFrame):
 		if max(size) >= max(display_client_size):
 			scale = 50
 		if debug: safe_print("[D]  measureframe_size:", size)
-		self.SetMaxSize((-1, -1))
-		self.SetMinSize(size)
-		self.SetSize(size)
-		self.SetMaxSize(size)
+		if (sys.platform not in ("darwin", "win32") and
+			os.getenv("XDG_SESSION_TYPE") != "wayland" and
+			not os.getenv("XDG_CURRENT_DESKTOP", "").startswith("KDE") and
+			hasattr(self, "MaxClientSize")):
+			self.MaxClientSize = (-1, -1)
+			self.MinClientSize = size
+			self.ClientSize = size
+			self.MaxClientSize = size
+		else:
+			self.SetMaxSize((-1, -1))
+			self.SetMinSize(size)
+			self.SetSize(size)
+			self.SetMaxSize(size)
 		display_rect = display[1]
 		if debug: safe_print("[D]  display_rect:", display_rect)
 		display_size = display_rect[2:]
@@ -399,14 +463,18 @@ class MeasureFrame(InvincibleFrame):
 
 	def close_handler(self, event):
 		if debug: safe_print("[D] measureframe_close_handler")
-		self.Hide()
 		if self.Parent:
+			if self.Parent.worker.is_working():
+				self.Parent.worker.abort_subprocess(confirm=True)
+				return
+			self.Hide()
 			self.Parent.Show()
 			if getattr(self.Parent, "restore_measurement_mode"):
 				self.Parent.restore_measurement_mode()
 			if getattr(self.Parent, "restore_testchart"):
 				self.Parent.restore_testchart()
 		else:
+			self.Hide()
 			writecfg()
 			self.Destroy()
 			if MeasureFrame.exitcode != 255:
@@ -467,6 +535,8 @@ class MeasureFrame(InvincibleFrame):
 		display_no, geometry, client_area = self.get_display()
 		if display_no != self.display_no:
 			self.display_no = display_no
+			if config.is_virtual_display():
+				return
 			# Translate from wx display index to Argyll display index
 			n = get_argyll_display_number(geometry)
 			if n is not None:
@@ -475,9 +545,74 @@ class MeasureFrame(InvincibleFrame):
 
 	def focus_handler(self, e):
 		e.Skip()
-		if e.EventObject == self:
-			# Need to give focus to the panel so controls can gain focus
-			self.panel.SetFocus()
+		if debug:
+			safe_print("SET_FOCUS", e.EventObject.Name)
+		if e.EventObject is self and getattr(self, "last_focused",
+											 None) not in (None, self):
+			self.last_focused.SetFocus()
+			if debug:
+				safe_print(self.last_focused.Name + ".SetFocus()")
+
+	def focus_lost_handler(self, e):
+		e.Skip()
+		if debug:
+			safe_print("KILL_FOCUS", e.EventObject.Name)
+		if e.EventObject is not self:
+			self.last_focused = e.EventObject
+			if debug and self.last_focused:
+				safe_print("last_focused", self.last_focused.Name)
+
+	def show_handler(self, e):
+		e.Skip()
+		if getattr(e, "IsShown", getattr(e, "GetShow", bool))():
+			self.measurebutton.SetFocus()
+
+	def show_controls(self, show=True):
+		self.panel.Freeze()
+		for ctrl in self.panel.Children:
+			ctrl.Show(show)
+		if show:
+			self.panel.SetBitmap(None)
+			self.panel.SetCursor(wx.StockCursor(wx.CURSOR_DEFAULT))
+		else:
+			self.panel.SetCursor(wx.StockCursor(wx.CURSOR_BLANK))
+		self.panel.Layout()
+		self.panel.Thaw()
+
+	def show_rgb(self, rgb):
+		rgb = tuple(v * 255 for v in rgb)
+		floor = tuple(int(math.floor(v)) for v in rgb)
+		ceil = tuple(int(math.ceil(v)) for v in rgb)
+		if floor != ceil:
+			# Dither using simple ordered pattern
+			safe_print("Dither 8 bit %.6f %.6f %.6f -> %i %i %i | %i %i %i" %
+					   (rgb + floor + ceil))
+			img = wx.EmptyImage(*self.ClientSize, clear=False)
+			buf = img.GetDataBuffer()
+			buflen = len(buf)
+			# Intervals in pixels per each R, G and B
+			intervals = tuple((buflen / (buflen * (rgb[i] - floor[i]))
+							  if rgb[i] - floor[i] else 0) for i in xrange(3))
+			safe_print("Intervals %.6f %.6f %.6f" % intervals)
+			floorbytes = tuple(chr(v) for v in floor)
+			ceilbytes = tuple(chr(v) for v in ceil)
+			n = 0
+			for i, byte in enumerate(buf):
+				m = intervals[i % 3]
+				if m and n % m < 1:
+					color = ceilbytes
+				else:
+					color = floorbytes
+				buf[i] = color[i % 3]
+				if i % 3 == 2:
+					n += 1
+			bmp = img.ConvertToBitmap()
+		else:
+			# Exact
+			safe_print("Exact 8 bit %.6f %.6f %.6f" % rgb)
+			bmp = wx.EmptyBitmapRGBA(*tuple(self.ClientSize) + floor, alpha=255)
+		self.panel.SetBitmap(bmp)
+		self.panel.Refresh()
 
 	def get_dimensions(self):
 		"""
@@ -543,12 +678,31 @@ class MeasureFrame(InvincibleFrame):
 		return measureframe_dimensions
 
 
+def test():
+	import time
+	for rgb in [(0.079291, 1 / 51., 1 / 51.),
+				(0.079291, 0.089572, 0.094845),
+				(0.032927, 0.028376, 0.027248),
+				(0.037647, 0.037095, 0.036181),
+				(51.2 / 255, 153.7 / 255, 127.4 / 255)]:
+		wx.CallAfter(wx.GetApp().TopWindow.show_rgb, rgb)
+		time.sleep(0.05)
+		raw_input("Press RETURN to continue\n")
+		if not wx.GetApp().TopWindow:
+			break
+
+
 def main():
 	config.initcfg()
 	lang.init()
 	app = BaseApp(0)
 	app.TopWindow = MeasureFrame()
 	app.TopWindow.Show()
+	if "--test-dither" in sys.argv[1:]:
+		import threading
+		t = threading.Thread(target=test)
+		app.TopWindow.show_controls(False)
+		t.start()
 	app.MainLoop()
 
 if __name__ == "__main__":
